'''
Bits and Code for Raster and DEM processing
'''

# Clipping rasters
from collections import namedtuple
#import fiona
import os
import rasterio
import rasterio.mask
import richdem as rd
import numpy as np
import pandas as pd
import cv2

def to_raster(Z, fname='example.tif', epsg='EPSG:3844', geoTransform=None):
    """
    [summary]

    :param Z: data to store as raster, 2D array
    :param fname: path/filename for raster, defaults to 'example.tif',  str, optional
    :param epsg: projection EPSG code, defaults to 'EPSG:3844',  str, optional
    :param geoTransform: raster geotransform, defaults to None,  array, optional

    TODO:
        - add support for 3D array / multiband raster
    """                          

    with rasterio.open(fname,
                        'w',
                        driver='GTiff',
                        height=Z.shape[0],
                        width=Z.shape[1],
                        count=1,
                        dtype=Z.dtype,
                        crs=epsg,
                        transform=geoTransform,
    ) as dst:
        dst.write(Z, 1)

def clip_raster(f_input, f_shp, f_output=None):
    '''
    Function to clip raster based on polygons
    
    :param f_input: Path to input raster file
    :param f_shp: Path to input shapefile containing polygons to clip raster with
    :param f_output: Path to output raster file 
    '''
    if f_output is None:
        f_output = f_input[:-4] + '_clip.tif'
        
    with fiona.open(f_shp, "r") as shapefile:
        shapes = [feature["geometry"] for feature in shapefile]
        
    with rasterio.open(f_input) as src:
        out_image, out_transform = rasterio.mask.mask(src, shapes, crop=True)
        out_meta = src.meta
        
    out_meta.update({"driver": "GTiff",
                 "height": out_image.shape[1],
                 "width": out_image.shape[2],
                 "transform": out_transform})
    
    with rasterio.open(f_output, "w", **out_meta) as dest:
        dest.write(out_image)
        print("Clipped raster stored at: ")
        print(f_output)
        print('¨¨¨¨¨¨¨')
        
def lapogaus(x, y, sigma):
    '''
    Laplacian of Gaussian operator for curvature computation smoothing noise.
    https://homepages.inf.ed.ac.uk/rbf/HIPR2/log.htm
    
    :param x: array of x-coordinate. Should be centered on 0.
    :param y: array of y-coordinate. Should be centered on 0.
    :param sigma: float, standard deviation setting the with of the bell curve
    :return: array containing the operator
    '''
    op = -1/(np.pi * sigma **4)*(1-((x**2 + y**2)/(2*sigma**2)))*np.exp(-((x**2 + y**2)/(2*sigma**2)))
    return op

def laplacian_of_gaussian(n, sigma):
    x = np.arange(-n,n)
    y = np.arange(-n,n)
    Xs, Ys = np.meshgrid(x,y)
    kernel = lapogaus(Xs,Ys,sigma)
    #print('Kernel sum = ',np.sum(kernel))
    return kernel

def plot_kernel(kernel):
    plt.figure()
    plt.imshow(kernel)
    plt.colorbar()

def convolve_array(arr, kernel):
    res = cv2.filter2D(arr, -1, kernel)
    return res
        
        
# Derive curvature from DEM
def compute_curvatures(f_input, no_data=-9999, method='LoG', kernel_size=71, sigma=15):
    if method == 'LoG':
        rast = rasterio.open(f_input)
        arr = rast.read(1)
        
        kernel = laplacian_of_gaussian(kernel_size, sigma)
        print('Kernel {}*{} with $\sigma$={}'.format(kernel_size, kernel_size, sigma))
        #plot_kernel(kernel)
        curv = cv2.filter2D(arr, -1, kernel)
        fname = f_input[:-4] + '_curv.tif'
        with rasterio.open(fname, 'w', **rast.profile) as  dst:
            dst.write(curv, 1)
            print("Raster exported: ", fname)
            
    if method == 'richdem':
        dem = rd.LoadGDAL(f_input, no_data=no_data)
        curv_profile = rd.TerrainAttribute(dem, attrib='profile_curvature')
        curv_planform = rd.TerrainAttribute(dem, attrib='planform_curvature')
        rd.SaveGDAL(f_input[:-4]+'_curv_prof.tif', curv_profile)
        rd.SaveGDAL(f_input[:-4]+'_curv_plan.tif', curv_planform)
        print("Curvature rasters stored at: ")
        print(f_input[:-4]+'_curv_prof.tif')
        print(f_input[:-4]+'_curv_plan.tif')
        print('¨¨¨¨¨¨¨')
    
    
def sample_raster(df, f_input, col_name='value', band=1):
    with rasterio.open(f_input) as raster:
        geot = raster.get_transform()
        x = (df["X"] - geot[0])/geot[1]
        y = (df["Y"] - geot[3])/geot[-1]
        df[col_name] = raster.read(band)[y.astype(np.int), x.astype(np.int)]
        df[col_name].loc[df[col_name]<-10**30] = np.nan
    return df



class massif():
    def __init__(self, name, fullname=None, pwd=None):
        self.name = name
        self.fullname = fullname
        self.extent = None
        self.dem = None
        self.slope = None
        self.mask = None
        self.ndvi = None
        self.solar = None
        self.aspect = None
        self.curv = None
        self.curv_plan = None
        self.curv_profile = None
        self.epsg = None
        self.geot = None
        self.snow_prob_2017 = None
        self.snow_prob_2018 = None
        self.snow_prob_2019 = None
        self.snow_prob_2020 = None
        self.pwd = pwd
        self.diurnal_anisotropic_heat = None
        self.raster_dict = None
        self.thresholds = None
        self.models = None
        
    def _update_object(self):
        self.dem = self.raster_dict['dem']
        self.mask = self.raster_dict['mask']
        self.ndvi = self.raster_dict['ndvi']
        self.solar = self.raster_dict['solar']
        self.aspect = self.raster_dict['aspect']
        #self.diurnal_anisotropic_heat = self.raster_dict['diurnal_anisotropic_heat']

        #self.curv_profile = self.raster_dict['curv_profile']
        #self.curv_plan = self.raster_dict['curv_planform']
        self.curv = self.raster_dict['curv']
        self.snow_prob_2017 = self.raster_dict['snow_prob_2017']
        self.snow_prob_2018 = self.raster_dict['snow_prob_2018']
        self.snow_prob_2019 = self.raster_dict['snow_prob_2019']
        self.snow_prob_2020 = self.raster_dict['snow_prob_2020']

    def _update_raster_dict(self):
        self.raster_dict = {
            'dem':self.dem,
            'mask':self.mask,
            'ndvi':self.ndvi,
            'slope':self.slope,
            'solar':self.solar,
            'aspect':self.aspect,
            #'green':self.green,
            'curv': self.curv,
            #'curv_planform': self.curv_plan,
            #'curv_profile': self.curv_profile,
            #'diurnal_anisotropic_heat': self.diurnal_anisotropic_heat,
            'snow_prob_median':self.snow_prob_median,
            #'snow_prob_std':self.snow_prob_std,
            #'snow_prob_2017': self.snow_prob_2017,
            #'snow_prob_2018': self.snow_prob_2018,
            #'snow_prob_2019': self.snow_prob_2019,
            #'snow_prob_2020': self.snow_prob_2020,
        }


    def to_df(self, var=None):
        self.df = pd.DataFrame()
        if var is None:
            var = self.raster_dict
        for val in var:
            with rasterio.open(self.raster_dict[val]) as src:
                self.df[val] = src.read(1).flatten()
        xs = np.arange(src.get_transform()[0], 
                    src.get_transform()[0]+src.width*src.get_transform()[1],
                    src.get_transform()[1])
        ys = np.arange(src.get_transform()[3]+src.height*src.get_transform()[5],
                    src.get_transform()[3], 
                    -src.get_transform()[5])
        X, Y = np.meshgrid(xs,ys)
        self.df['X'] = X.flatten()
        self.df['Y'] = Y.flatten()
        

    def clip_raster_cmd(self, f_input=None, f_output=None, epsg_src=None, epsg_tgt=None, resolution=None, extent=None, execute=False):
        """
        [summary]

        :param f_input: [description], defaults to None,  [type], optional
        :param f_output: [description], defaults to None,  [type], optional
        :param epsg_src: [description], defaults to None,  [type], optional
        :param epsg_tgt: [description], defaults to None,  [type], optional
        :param resolution: [description], defaults to None,  [type], optional
        :param extent: [description], defaults to None,  [type], optional
        :param execute: [description], defaults to False,  bool, optional
        :return: [description], [type]
        """              
        if epsg_src is None:
            epsg_src = self.epsg
        if epsg_tgt is None:
            epsg_tgt = self.epsg
        if extent is None:
            extent = self.extent
        if resolution is None:
            resolution = self.resolution
        if f_output is None:
            f_output = f_input[:-4] + '_clip.tif'

        cmd = "gdalwarp -s_srs {} -t_srs {} -te {} {} {} {} -te_srs {} -tr {} {} -r bilinear '{}' '{}'".format(
            epsg_src,
            epsg_tgt,
            extent[0],
            extent[1],
            extent[2],
            extent[3],
            epsg_tgt,
            resolution,
            resolution,
            f_input,
            f_output

        )
        print(cmd)
        if execute:
            os.system(cmd)

        return cmd

    def clip_rasters(self, epsg_tgt='EPSG:3844', resolution=10, execute=True):
        self._update_raster_dict()
        for var in self.raster_dict:
            if self.raster_dict[var] is not None:
                with rasterio.open(self.raster_dict[var], 'r') as rast:
                    epsg = rast.crs.data['init']
                    if (self.raster_dict[var][-9:]) != '_clip.tif':
                        if var[:4] == 'snow':
                            fout = self.pwd + self.name + '_' + var + '_clip.tif'
                        else:
                            fout = self.pwd + self.raster_dict[var].split('/')[-1][:-4] + '_clip.tif'
                    else:
                        fout = self.pwd + self.raster_dict[var].split('/')[-1]
                    if execute:
                        self.clip_raster_cmd(self.raster_dict[var], fout, epsg_src=epsg, epsg_tgt=epsg_tgt, resolution=resolution, extent=self.extent, execute=True)
                    self.raster_dict[var] = fout
        self._update_object()
        with rasterio.open(self.dem, 'r') as rast:
            self.width = rast.width
            self.height = rast.height
            self.geot = rast.get_transform()
            self.bounds = rast.bounds
            self.raster_profile = rast.profile
            self.raster_profile.pop('dtype')

    def get_rasterio_meta(self):
        with rasterio.open(self.dem, 'r') as rast:
            self.width = rast.width
            self.height = rast.height
            self.geot = rast.get_transform()
            self.bounds = rast.bounds
            self.raster_profile = rast.profile
            self.raster_profile.pop('dtype')

    def compute_curvatures(self, execute=True):
        if execute:
            beau = rd.LoadGDAL(self.dem)
            planform_curvature = rd.TerrainAttribute(beau, attrib='planform_curvature')
            profile_curvature = rd.TerrainAttribute(beau, attrib='profile_curvature')
        fname_plan = self.dem[:-4] + '_curve_plan.tif'
        fname_prof = self.dem[:-4] + '_curve_prof.tif'
        if execute:
            rd.SaveGDAL(fname_prof, profile_curvature)
            rd.SaveGDAL(fname_plan, planform_curvature)
        self.curv_profile = fname_prof
        self.curv_plan = fname_plan
        self._update_raster_dict()

    def compute_diurnal_anisotropic_heat(self, execute=False):
        """
        Function to compute diurnal anisotropic heat using SAGA routine
        http://www.saga-gis.org/saga_tool_doc/7.9.0/ta_morphometry_12.html

        :param execute: flag to run command, defaults to False,  bool, optional

        TODO:
        - add remove SAGA files after gdal_translate
        """        
        self.diurnal_anisotropic_heat = self.dem[:-4] + '_diurnalHeat.sdat'
        cmd1 = "saga_cmd ta_morphometry 12 -DEM '{}' -DAH '{}' -ALPHA_MAX {}".format(self.dem, self.diurnal_anisotropic_heat, '202.5')
        print(cmd1)

        # translating default raster format '.sdat' generated by SAGA to geotiff
        cmd2 = "gdal_translate '{}' '{}'".format(self.diurnal_anisotropic_heat, self.diurnal_anisotropic_heat[:-5]+'.tif')
        self.diurnal_anisotropic_heat = self.diurnal_anisotropic_heat[:-5]+'.tif'
        print(cmd2)

        if execute:
            os.system(cmd1)
            os.system[cmd2]
        self._update_raster_dict()

    def compute_snow_stats(self, execute=True):
        stack = np.zeros((int(self.height),int(self.width), 4))
        i=0
        for key in self.raster_dict:
            if key.startswith('snow'):
                print(self.raster_dict[key])
                stack[:,:,i] = rasterio.open(self.raster_dict[key]).read(1)
                i+=1
        self.snow_prob_median = self.pwd + self.name + '_snow_median.tif'
        self.snow_prob_std = self.pwd + self.name + '_snow_std.tif'

        if execute:
            with rasterio.open(self.snow_prob_median, 'w', dtype=stack.dtype, **self.raster_profile) as  dst:
                dst.write(np.median(stack, axis=2), 1)
            with rasterio.open(self.snow_prob_std, 'w', dtype=stack.dtype, **self.raster_profile) as  dst:
                dst.write(np.std(stack, axis=2), 1)
            
        self.raster_dict.update({'snow_prob_median':self.snow_prob_median, 'snow_prob_std':self.snow_prob_std})


    def map_to_raster(self, model="Logistic Regression", fname=None):
        """
        Function to export map as raster
        
        :param model: name of the DataFrame column to export as raster
        :pram fname:
        """
        arr = self.df[model].values.reshape(self.height, self.width)
        if fname is None:
            fname = self.pwd + self.name + '_prob_'+ model.replace(' ', '_') + '.tif'

        with rasterio.open(fname, 'w', dtype=arr.dtype, **self.raster_profile) as  dst:
            dst.write(arr, 1)
            print("Raster exported: ", fname)
